import djangoevents
from unittest import mock


@mock.patch.object(djangoevents, 'is_validation_enabled', return_value=True)
@mock.patch.object(djangoevents, 'validate_event', return_value=True)
@mock.patch.object(djangoevents, 'es_publish', return_value=True)
def test_store_event_validation_enabled(publish, validate_event, validation_enabled):
    evt = {}
    djangoevents.store_event(evt)

    assert validation_enabled.call_count == 1
    assert validate_event.call_args_list == [mock.call(evt)]
    assert publish.call_args_list == [mock.call(evt)]


@mock.patch.object(djangoevents, 'is_validation_enabled', return_value=False)
@mock.patch.object(djangoevents, 'validate_event', return_value=True)
@mock.patch.object(djangoevents, 'es_publish', return_value=True)
def test_store_event_validation_disabled(publish, validate_event, validation_enabled):
    evt = {}
    djangoevents.store_event(evt)

    assert validation_enabled.call_count == 1
    assert validate_event.call_count == 0
    assert publish.call_args_list == [mock.call(evt)]


@mock.patch.object(djangoevents, 'is_validation_enabled', return_value=False)
@mock.patch.object(djangoevents, 'validate_event', return_value=True)
@mock.patch.object(djangoevents, 'es_publish', return_value=True)
def test_store_event_validation_disabled_force(publish, validate_event, validation_enabled):
    evt = {}
    djangoevents.store_event(evt, force_validate=True)

    assert validation_enabled.call_count == 1
    assert validate_event.call_args_list == [mock.call(evt)]
    assert publish.call_args_list == [mock.call(evt)]



from ..domain import BaseEntity, DomainEvent
from ..schema import get_event_version
from ..schema import set_event_version
from ..unifiedtranscoder import UnifiedTranscoder
from django.core.serializers.json import DjangoJSONEncoder
from django.test import override_settings
from unittest import mock

import os
import pytest


class SampleEntity(BaseEntity):
    class Created(BaseEntity.Created):
        def mutate_event(self, event, klass):
            return klass(entity_id=event.entity_id,
                         entity_version=event.entity_version,
                         domain_event_id=event.domain_event_id)

    class Updated(DomainEvent):
        def mutate_event(self, event, entity):
            entity.is_dirty = True
            return entity

    class Closed(DomainEvent):
        # no mutate_event
        pass

    def __init__(self, is_dirty=False, **kwargs):
        super().__init__(**kwargs)
        self.is_dirty = is_dirty


def test_base_entity_calls_mutator():
    create_event = SampleEntity.Created(entity_id=1)
    entity = SampleEntity.mutate(event=create_event)

    close_event = SampleEntity.Updated(entity_id=entity.id, entity_version=entity.version)
    entity = SampleEntity.mutate(event=close_event, entity=entity)
    assert entity.is_dirty is True


def test_base_entity_requires_mutator():
    create_event = SampleEntity.Created(entity_id=1)
    entity = SampleEntity.mutate(event=create_event)

    close_event = SampleEntity.Closed(entity_id=entity.id, entity_version=entity.version)

    with pytest.raises(NotImplementedError):
        SampleEntity.mutate(event=close_event, entity=entity)


def test_create_for_event():
    event = SampleEntity.Created(
        entity_id='ENTITY_ID',
        domain_event_id='DOMAIN_EVENT_ID',
        entity_version=0,
    )
    obj = SampleEntity.create_for_event(event)

    assert obj.id == 'ENTITY_ID'
    assert obj.version == 0


@override_settings(BASE_DIR='/path/to/proj/src/')
def test_version_1():
    assert get_event_version(SampleEntity.Created) == 1


def test_version_4(tmpdir):
    # make temporary directory structure
    avro_dir = tmpdir.mkdir('avro_dir')
    entity_dir = avro_dir.mkdir('sample_entity')

    original_version = getattr(SampleEntity.Created, 'version', None)
    try:
        for version in range(1, 4):
            # make empty schema file
            expected_schema_path = os.path.join(entity_dir.strpath, 'v{}_sample_entity_created.json'.format(version))
            with open(expected_schema_path, 'w'):
                pass

            # refresh version
            set_event_version(SampleEntity, SampleEntity.Created, avro_dir=avro_dir.strpath)

            assert get_event_version(SampleEntity.Created) == version
    finally:
        SampleEntity.Created.version = original_version


@override_settings(DJANGOEVENTS_CONFIG={
    'ADDS_SCHEMA_VERSION_TO_EVENT_DATA': False,
})
def test_events_dont_have_schema_version_when_disabled():
    event = SampleEntity.Created(entity_id=1)
    assert not hasattr(event, 'schema_version')


@override_settings(DJANGOEVENTS_CONFIG={
    'ADDS_SCHEMA_VERSION_TO_EVENT_DATA': True,
})
def test_events_have_schema_version_when_enabled():
    event = SampleEntity.Created(entity_id=1)
    assert event.schema_version == 1


@override_settings(DJANGOEVENTS_CONFIG={
    'ADDS_SCHEMA_VERSION_TO_EVENT_DATA': True,
})
def test_events_have_correct_schema_version():
    with mock.patch.object(SampleEntity.Created, 'version', 666):
        event = SampleEntity.Created(entity_id=1)
        assert event.schema_version == 666


@override_settings(DJANGOEVENTS_CONFIG={
    'ADDS_SCHEMA_VERSION_TO_EVENT_DATA': True,
})
def test_explicitly_provided_schema_version_is_not_overridden():
    with mock.patch.object(SampleEntity.Created, 'version', 2):
        event = SampleEntity.Created(entity_id=1, schema_version=3)
        assert event.schema_version == 3


@override_settings(DJANGOEVENTS_CONFIG={
    'ADDS_SCHEMA_VERSION_TO_EVENT_DATA': True,
})
def test_schema_version_is_not_overridden_when_serialized_and_deserialized():
    transcoder = UnifiedTranscoder(json_encoder_cls=DjangoJSONEncoder)

    with mock.patch.object(SampleEntity.Created, 'version', 2):
        event = SampleEntity.Created(entity_id=1)

    stored_event = transcoder.serialize(event)

    with mock.patch.object(SampleEntity.Created, 'version', 3):
        entity = transcoder.deserialize(stored_event)

    assert entity.schema_version == 2



import inspect
from .domain import BaseEntity
from .domain import BaseAggregate
from .domain import DomainEvent
from djangoevents.utils_abstract import is_abstract
import re


def list_concrete_aggregates():
    """
    Lists all non abstract aggregates defined within the application.
    """
    aggregates = set(_list_subclasses(BaseAggregate) + _list_subclasses(BaseEntity))
    return [aggregate for aggregate in aggregates if not aggregate.is_abstract_class()]


def is_event_mutating(event):
    return hasattr(event, 'mutate_event')


def list_aggregate_events(aggregate_cls):
    """
    Lists all aggregate_cls events defined within the application.
    Note: Only events with a defined `mutate_event` flow and are not marked as abstract will be returned.
    """
    events = _list_internal_classes(aggregate_cls, DomainEvent)
    return [event_cls for event_cls in events if is_event_mutating(event_cls) and not is_abstract(event_cls)]


def event_to_json(event):
    """
    Converts an event class to its dictionary representation.
    Underlying eventsourcing library does not provide a proper event->dict conversion function.

    Note: Similarly to event journal persistence flow, this method supports native JSON types only.
    """
    return vars(event)


def _list_subclasses(cls):
    """
    Recursively lists all subclasses of `cls`.
    """
    subclasses = cls.__subclasses__()

    for subclass in cls.__subclasses__():
        subclasses += _list_subclasses(subclass)

    return subclasses


def _list_internal_classes(cls, base_class=None):
    base_class = base_class or object

    return [cls_attribute for cls_attribute in cls.__dict__.values()
            if inspect.isclass(cls_attribute)
            and issubclass(cls_attribute, base_class)]


def camel_case_to_snake_case(text):

    def repl(match):
        sep = match.group().lower()
        if match.start() > 0:
            sep = '_%s' % sep
        return sep

    return re.sub(r'[A-Z][a-z]', repl, text).lower()



import pytest
from ..unifiedtranscoder import UnifiedTranscoder
from eventsourcing.domain.model.entity import EventSourcedEntity
from django.core.serializers.json import DjangoJSONEncoder
from django.test.utils import override_settings
from unittest import mock


class SampleAggregate(EventSourcedEntity):
    class Created(EventSourcedEntity.Created):
        pass

    class Updated(EventSourcedEntity.AttributeChanged):
        event_type = 'overridden_event_type'


def test_serialize_and_deserialize_1():
    transcoder = UnifiedTranscoder(json_encoder_cls=DjangoJSONEncoder)
    # test serialize
    created = SampleAggregate.Created(entity_id='b089a0a6-e0b3-480d-9382-c47f99103b3d', attr1='val1', attr2='val2',
                                      metadata={'command_id': 123})
    created_stored_event = transcoder.serialize(created)
    assert created_stored_event.event_type == 'sample_aggregate_created'
    assert created_stored_event.event_version == 1
    assert created_stored_event.event_data == '{"attr1":"val1","attr2":"val2"}'
    assert created_stored_event.aggregate_id == 'b089a0a6-e0b3-480d-9382-c47f99103b3d'
    assert created_stored_event.aggregate_version == 0
    assert created_stored_event.aggregate_type == 'SampleAggregate'
    assert created_stored_event.module_name == 'djangoevents.tests.test_unifiedtranscoder'
    assert created_stored_event.class_name == 'SampleAggregate.Created'
    assert created_stored_event.metadata == '{"command_id":123}'
    # test deserialize
    created_copy = transcoder.deserialize(created_stored_event)
    assert 'metadata' not in created_copy.__dict__
    created.__dict__.pop('metadata')  # metadata is not included in deserialization
    assert created.__dict__ == created_copy.__dict__


def test_serialize_and_deserialize_2():
    transcoder = UnifiedTranscoder(json_encoder_cls=DjangoJSONEncoder)
    # test serialize
    updated = SampleAggregate.Updated(entity_id='b089a0a6-e0b3-480d-9382-c47f99103b3d', entity_version=10, attr1='val1',
                                      attr2='val2', metadata={'command_id': 123})
    updated_stored_event = transcoder.serialize(updated)
    assert updated_stored_event.event_type == 'overridden_event_type'
    assert updated_stored_event.event_version == 1
    assert updated_stored_event.event_data == '{"attr1":"val1","attr2":"val2"}'
    assert updated_stored_event.aggregate_id == 'b089a0a6-e0b3-480d-9382-c47f99103b3d'
    assert updated_stored_event.aggregate_version == 10
    assert updated_stored_event.aggregate_type == 'SampleAggregate'
    assert updated_stored_event.module_name == 'djangoevents.tests.test_unifiedtranscoder'
    assert updated_stored_event.class_name == 'SampleAggregate.Updated'
    assert updated_stored_event.metadata == '{"command_id":123}'
    # test deserialize
    updated_copy = transcoder.deserialize(updated_stored_event)
    assert 'metadata' not in updated_copy.__dict__
    updated.__dict__.pop('metadata') # metadata is not included in deserialization
    assert updated.__dict__ == updated_copy.__dict__


@override_settings(DJANGOEVENTS_CONFIG={
    'ADDS_SCHEMA_VERSION_TO_EVENT_DATA': False,
})
def test_serializer_doesnt_include_schema_version_when_its_disabled():
    transcoder = UnifiedTranscoder(json_encoder_cls=DjangoJSONEncoder)
    event = SampleAggregate.Created(entity_id='b089a0a6-e0b3-480d-9382-c47f99103b3d', attr1='val1', attr2='val2')
    serialized_event = transcoder.serialize(event)
    assert serialized_event.event_data == '{"attr1":"val1","attr2":"val2"}'


@override_settings(DJANGOEVENTS_CONFIG={
    'ADDS_SCHEMA_VERSION_TO_EVENT_DATA': True,
})
def test_serializer_includes_schema_version_when_its_enabled():
    transcoder = UnifiedTranscoder(json_encoder_cls=DjangoJSONEncoder)
    event = SampleAggregate.Created(entity_id='b089a0a6-e0b3-480d-9382-c47f99103b3d', attr1='val1', attr2='val2')
    serialized_event = transcoder.serialize(event)
    assert serialized_event.event_data == '{"attr1":"val1","attr2":"val2","schema_version":1}'


def test_metadata_is_optional():
    transcoder = UnifiedTranscoder(json_encoder_cls=DjangoJSONEncoder)
    created = SampleAggregate.Created(entity_id='b089a0a6-e0b3-480d-9382-c47f99103b3d')

    try:
        transcoder.serialize(created)
    except AttributeError:
        pytest.fail('Serialization of an event without metadata should not raise an exception.')


def test_transcoder_passes_all_attributes_to_event_constructor():
    attributes = {
        'entity_id': 'b089a0a6-e0b3-480d-9382-c47f99103b3d',
        'foo': 0,
        'bar': 1,
    }

    event = SampleAggregate.Created(**attributes)
    transcoder = UnifiedTranscoder(json_encoder_cls=DjangoJSONEncoder)
    serialized_event = transcoder.serialize(event)

    init_call_args = None
    def init(self, *args, **kwargs):
        nonlocal init_call_args
        init_call_args = (args, kwargs)

    with mock.patch.object(SampleAggregate.Created, '__init__', init):
        transcoder.deserialize(serialized_event)

    args, kwargs = init_call_args
    assert args == tuple()

    for key, value in attributes.items():
        assert key in kwargs
        assert kwargs[key] == value



from djangoevents.domain import BaseAggregate
from djangoevents.domain import DomainEvent
from djangoevents.utils_abstract import abstract
from ..utils import camel_case_to_snake_case
from ..utils import list_aggregate_events
from ..utils import list_concrete_aggregates
from ..utils import _list_subclasses
from ..utils import _list_internal_classes
from unittest import mock
import pytest


def test_list_subclasses():
    class Parent:
        pass

    class Child(Parent):
        pass

    class GrandChild(Child):
        pass

    assert set(_list_subclasses(Parent)) == {Child, GrandChild}


def test_list_subclasses_when_none():
    class Parent:
        pass

    assert _list_subclasses(Parent) == []


def list_internal_classes():
    class Parent:
        class Child1:
            pass

        class Child2:
            class GrandChild:
                pass

    # Please note that GrandChild is not on the list!
    assert set(_list_internal_classes(Parent)) == {Parent.Child1, Parent.Child2}


def list_internal_classes_none():
    class Parent:
        pass

    assert _list_internal_classes(Parent) == []


def test_list_aggregates_skip_abstract():
    class Aggregate1(BaseAggregate):
        pass

    @abstract
    class Aggregate2(BaseAggregate):
        pass

    with mock.patch('djangoevents.utils._list_subclasses') as list_subclasses:
        list_subclasses.return_value = [Aggregate1, Aggregate2]
        aggregates = list_concrete_aggregates()
        assert aggregates == [Aggregate1]


def test_list_aggregates_none_present():
    with mock.patch('djangoevents.utils._list_subclasses') as list_subclasses:
        list_subclasses.return_value = []
        aggregates = list_concrete_aggregates()
        assert aggregates == []


def test_list_events_sample_event_appart_from_abstract():
    class Aggregate(BaseAggregate):
        class Evt1(DomainEvent):
            def mutate_event(self, *args, **kwargs):
                pass

        class Evt2(DomainEvent):
            def mutate_event(self, *args, **kwargs):
                pass

        class Evt3(DomainEvent):
            # No mutate_event present
            pass

        @abstract
        class Evt4(DomainEvent):
            def mutate_event(self, *args, **kwargs):
                pass

    events = list_aggregate_events(Aggregate)
    assert set(events) == {Aggregate.Evt1, Aggregate.Evt2}


def test_list_events_not_an_aggregate():
    events = list_aggregate_events(list)
    assert events == []


@pytest.mark.parametrize('name, expected_output', [
    ('UserRegistered', 'user_registered'),
    ('UserRegisteredWithEmail', 'user_registered_with_email'),
    ('HttpResponse', 'http_response'),
    ('HTTPResponse', 'http_response'),
    ('already_snake', 'already_snake'),
])
def test_camel_case_to_snake_case(name, expected_output):
    assert expected_output == camel_case_to_snake_case(name)



from ..shortcuts import get_entity_or_404, get_aggregate_or_404
from django.http import Http404

import pytest


def test_get_entity_or_404_item_exists():
    """
    ES store provides a dictionary style item verification.
    """
    repo = {'id': 'entity'}
    assert get_entity_or_404(repo, 'id') == 'entity'


def test_get_entity_or_404_item_does_not_exists():
    repo = {}
    with pytest.raises(Http404) as exc_info:
        get_entity_or_404(repo, 'id')


def test_get_aggregate_or_404_item_exists():
    repo = {'id': 'aggregate'}
    assert get_aggregate_or_404(repo, 'id') == 'aggregate'


def test_get_aggregate_or_404_item_does_not_exists():
    repo = {}
    with pytest.raises(Http404) as exc_info:
        get_aggregate_or_404(repo, 'id')



import warnings

from eventsourcing.domain.model.entity import EventSourcedEntity
from eventsourcing.domain.model.entity import entity_mutator
from eventsourcing.domain.model.entity import singledispatch
from eventsourcing.domain.model.decorators import subscribe_to
from eventsourcing.domain.model.events import publish as es_publish
from eventsourcing.domain.model.events import subscribe
from eventsourcing.domain.model.events import unsubscribe
from eventsourcing.infrastructure.event_sourced_repo import EventSourcedRepository
from .domain import BaseEntity
from .domain import BaseAggregate
from .domain import DomainEvent
from .app import EventSourcingWithDjango
from .exceptions import EventSchemaError
from .schema import validate_event
from .settings import is_validation_enabled

default_app_config = 'djangoevents.apps.AppConfig'

__all__ = [
    'DomainEvent',
    'EventSourcedEntity',
    'EventSourcedRepository',
    'entity_mutator',
    'singledispatch',
    'publish',
    'store_event'
    'subscribe',
    'unsubscribe',
    'subscribe_to',
    'BaseEntity',
    'BaseAggregate',
    'EventSourcingWithDjango'
]


def publish(event):
    warnings.warn("`publish` is depreciated. Please switch to: `store_event`.", DeprecationWarning)
    return es_publish(event)


def store_event(event, force_validate=False):
    """
    Store an event to the service's event journal. Optionally validates event
    schema if one is provided.

    `force_validate` - enforces event schema validation even if configuration disables it globally.
    """
    if is_validation_enabled() or force_validate:
        is_valid = validate_event(event)
        if not is_valid:
            msg = "Event: {} does not match its schema.".format(event)
            raise EventSchemaError(msg)

    return es_publish(event)



from djangoevents.domain import BaseAggregate
from djangoevents.domain import DomainEvent
from djangoevents.utils_abstract import abstract
from djangoevents.utils_abstract import is_abstract


def test_subclass_of_abstract_event_is_not_abstract():

    class Aggregate(BaseAggregate):

        @abstract
        class Event1(DomainEvent):
            pass

        class Event2(Event1):
            pass

    assert is_abstract(Aggregate.Event1)
    assert not is_abstract(Aggregate.Event2)



import os
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


_DEFAULTS = {
    'EVENT_SCHEMA_VALIDATION': {
        'ENABLED': False,
        'SCHEMA_DIR': 'avro',
    },
    'ADDS_SCHEMA_VERSION_TO_EVENT_DATA': False,
}


def get_config():
    """
    Function getting djangoevents configuration defined in Django settings.
    We use a function because Django settings can be mutable (useful for example in tests).
    """
    return getattr(settings, 'DJANGOEVENTS_CONFIG', _DEFAULTS)


def is_validation_enabled():
    config = get_config()
    return config.get('EVENT_SCHEMA_VALIDATION', {}).get('ENABLED', False)


def get_avro_dir():
    config = get_config()

    try:
        avro_folder = config.get('EVENT_SCHEMA_VALIDATION', {})['SCHEMA_DIR']
    except KeyError:
        raise ImproperlyConfigured("Please define `SCHEMA_DIR`")

    avro_dir = os.path.abspath(os.path.join(settings.BASE_DIR, '..', avro_folder))
    return avro_dir


def adds_schema_version_to_event_data():
    config = get_config()
    return config.get('ADDS_SCHEMA_VERSION_TO_EVENT_DATA', False)



from django.db import models


class Event(models.Model):
    event_id = models.CharField(max_length=255, db_index=True)
    event_type = models.CharField(max_length=255)
    event_version = models.IntegerField(null=True)
    event_data = models.TextField()
    aggregate_id = models.CharField(max_length=255, db_index=True)
    aggregate_type = models.CharField(max_length=255)
    aggregate_version = models.IntegerField()
    create_date = models.DateTimeField()
    metadata = models.TextField()
    module_name = models.CharField(max_length=255, db_column='_module_name')
    class_name = models.CharField(max_length=255, db_column='_class_name')
    stored_entity_id = models.CharField(max_length=255, db_column='_stored_entity_id')

    class Meta:
        unique_together = (("aggregate_id", "aggregate_type", "aggregate_version"),)
        db_table = 'event_journal'

    def __str__(self):
        return '%s.%s | %s | %s' % (self.aggregate_type, self.event_type, self.event_id, self.create_date)



from django.apps import AppConfig as BaseAppConfig
from django.conf import settings
from django.utils.module_loading import import_module
from djangoevents import DomainEvent
from .exceptions import EventSchemaError
from .settings import adds_schema_version_to_event_data
from .schema import get_event_version
from .schema import load_all_event_schemas
import os.path
import warnings


class AppConfig(BaseAppConfig):
    name = 'djangoevents'

    def ready(self):
        patch_domain_event()

        for app_module_name in get_app_module_names():
            import_handlers_module(app_module_name)
            import_aggregates_module(app_module_name)

        # Once all handlers & aggregates are loaded we can import aggregate schema files.
        # `load_scheas()` assumes that all aggregates are imported at this point.
        load_schemas()


def get_app_module_names():
    return settings.INSTALLED_APPS


def import_handlers_module(app_module_name):
    return import_app_module(app_module_name, 'handlers')


def import_aggregates_module(app_module_name):
    return import_app_module(app_module_name, 'aggregates')


def import_app_module(app_module_name, module_name):
    full_module_name = '%s.%s' % (app_module_name, module_name)
    try:
        import_module(full_module_name)
    except ImportError:
        # we need to re-raise exception in case there was import error inside
        # `module_name` module
        module_file_name = get_module_file_name(app_module_name, module_name)
        if os.path.exists(module_file_name):
            raise

def get_module_file_name(app_module_name, module_name):
    module = import_module(app_module_name)
    module_dir = os.path.dirname(module.__file__)
    return os.path.join(module_dir, '%s.py' % module_name)


def load_schemas():
    """
    Try loading all the event schemas and complain loud if failure occurred.
    """
    try:
        load_all_event_schemas()
    except EventSchemaError as e:
        warnings.warn(str(e), UserWarning)


def patch_domain_event():
    """
    Patch `DomainEvent` to add `schema_version` to event payload.
    """

    old_init = DomainEvent.__init__

    def new_init(self, *args, **kwargs):
        old_init(self, *args, **kwargs)

        if adds_schema_version_to_event_data():
            dct = self.__dict__
            key = 'schema_version'
            if key not in dct:
                dct[key] = get_event_version(self.__class__)

    DomainEvent.__init__ = new_init



import random
import string
import django

DEBUG = False

SECRET_KEY = ''.join([random.choice(string.ascii_letters) for x in range(40)])


INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.admin',
    'django.contrib.messages',
    'djangoevents',
)

MIDDLEWARE = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)

if django.VERSION < (1, 10):
    MIDDLEWARE_CLASSES = MIDDLEWARE

TEST_RUNNER = 'django.test.runner.DiscoverRunner'

SITE_ID = 1


# Database specific

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

TEMPLATES = [
    {
    },
]



from .unifiedtranscoder import UnifiedStoredEvent
from django.conf import settings
from django.db.utils import IntegrityError
from django.utils import timezone
from djangoevents.exceptions import AlreadyExists
from eventsourcing.domain.services.eventstore import AbstractStoredEventRepository
from eventsourcing.domain.services.eventstore import EntityVersionDoesNotExist
from eventsourcing.domain.services.transcoding import EntityVersion
from eventsourcing.utils.time import timestamp_from_uuid
import datetime


class DjangoStoredEventRepository(AbstractStoredEventRepository):

    def __init__(self, *args, **kwargs):
        from .models import Event  # import model at runtime for making top level import possible
        self.EventModel = Event
        super(DjangoStoredEventRepository, self).__init__(*args, **kwargs)

    def append(self, new_stored_event, new_version_number=None, max_retries=3, artificial_failure_rate=0):
        """
        Saves given stored event in this repository.
        """
        # Check the new event is a stored event instance.
        assert isinstance(new_stored_event, UnifiedStoredEvent)

        # Not calling validate_expected_version!

        # Write the event.
        self.write_version_and_event(
            new_stored_event=new_stored_event,
            new_version_number=new_version_number,
            max_retries=max_retries,
            artificial_failure_rate=artificial_failure_rate,
        )

    def get_entity_version(self, stored_entity_id, version_number):
        events_query = self.EventModel.objects.filter(stored_entity_id=stored_entity_id)\
            .filter(aggregate_version=version_number)
        if not events_query.exists():
            raise EntityVersionDoesNotExist()

        return EntityVersion(
            entity_version_id=self.make_entity_version_id(stored_entity_id, version_number),
            event_id=events_query[0].event_id,
        )

    def write_version_and_event(self, new_stored_event, new_version_number=None, max_retries=3,
                                artificial_failure_rate=0):
        try:
            self.EventModel.objects.create(
                event_id=new_stored_event.event_id,
                event_type=new_stored_event.event_type,
                event_version=new_stored_event.event_version,
                event_data=new_stored_event.event_data,
                aggregate_id=new_stored_event.aggregate_id,
                aggregate_type=new_stored_event.aggregate_type,
                aggregate_version=new_stored_event.aggregate_version,
                create_date=make_aware_if_needed(new_stored_event.create_date),
                metadata=new_stored_event.metadata,
                module_name=new_stored_event.module_name,
                class_name=new_stored_event.class_name,
                stored_entity_id=new_stored_event.stored_entity_id,
            )
        except IntegrityError as err:
            create_attempt = not new_version_number
            if create_attempt:
                msg = "Aggregate with id %r already exists" % new_stored_event.aggregate_id
                raise AlreadyExists(msg)
            else:
                raise err

    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, query_ascending=True,
                          results_ascending=True):
        events = self.EventModel.objects.filter(stored_entity_id=stored_entity_id)
        if query_ascending:
            events = events.order_by('id')
        else:
            events = events.order_by('-id')

        if after is not None:
            after_ts = datetime.datetime.fromtimestamp(timestamp_from_uuid(after))
            if query_ascending:
                events = events.filter(create_date__gt=after_ts)
            else:
                events = events.filter(create_date__gte=after_ts)

        if until is not None:
            until_ts = datetime.datetime.fromtimestamp(timestamp_from_uuid(until))
            if query_ascending:
                events = events.filter(create_date__lte=until_ts)
            else:
                events = events.filter(create_date__lt=until_ts)

        if limit is not None:
            events = events[:limit]

        events = list(events)
        if results_ascending != query_ascending:
            events.reverse()
        return [from_model_instance(e) for e in events]


def from_model_instance(event):
    return UnifiedStoredEvent(
        event_id=event.event_id,
        event_type=event.event_type,
        event_version=event.event_version,
        event_data=event.event_data,
        aggregate_id=event.aggregate_id,
        aggregate_type=event.aggregate_type,
        aggregate_version=event.aggregate_version,
        create_date=event.create_date,
        metadata=event.metadata,
        module_name=event.module_name,
        class_name=event.class_name,
        stored_entity_id=event.stored_entity_id
    )


def make_aware_if_needed(dt):
    if settings.USE_TZ:
        return timezone.make_aware(dt)
    else:
        return dt



"""
Aggregate event avro schema validation
"""

import avro.schema
import itertools
import os
import stringcase

from avro.io import Validate as avro_validate
from .settings import get_avro_dir
from .utils import list_concrete_aggregates
from .utils import list_aggregate_events
from .utils import event_to_json
from .exceptions import EventSchemaError


schemas = {}


def load_all_event_schemas():
    """
    Initializes aggregate event schemas lookup cache.
    """
    errors = []
    for aggregate in list_concrete_aggregates():
        for event in list_aggregate_events(aggregate_cls=aggregate):
            try:
                schemas[event] = load_event_schema(aggregate, event)
            except EventSchemaError as e:
                errors.append(str(e))

    # Serve all schema errors at once not iteratively.
    if errors:
        raise EventSchemaError("\n".join(errors))

    return schemas


def set_event_version(aggregate_cls, event_cls, avro_dir=None):
    avro_dir = avro_dir or get_avro_dir()
    event_cls.version = _find_highest_event_version_based_on_schemas(aggregate_cls, event_cls, avro_dir)


def get_event_version(event_cls):
    return getattr(event_cls, 'version', None) or 1


def load_event_schema(aggregate, event):
    set_event_version(aggregate, event)
    spec_path = event_to_schema_path(aggregate, event)

    try:
        with open(spec_path) as fp:
            return parse_event_schema(fp.read())
    except FileNotFoundError:
        msg = "No event schema found for: {event} (expecting file at:{path})."
        raise EventSchemaError(msg.format(event=event, path=spec_path))
    except avro.schema.SchemaParseException as e:
        msg = "Can't parse schema for event: {event} from {path}."
        raise EventSchemaError(msg.format(event=event, path=spec_path)) from e


def event_to_schema_path(aggregate_cls, event_cls, avro_dir=None):
    avro_dir = avro_dir or get_avro_dir()
    version = get_event_version(event_cls)
    return _event_to_schema_path(aggregate_cls, event_cls, avro_dir, version)


def _find_highest_event_version_based_on_schemas(aggregate_cls, event_cls, avro_dir):
    version = None
    for version_candidate in itertools.count(1):
        schema_path = _event_to_schema_path(aggregate_cls, event_cls, avro_dir, version_candidate)
        if os.path.exists(schema_path):
            version = version_candidate
        else:
            break

    return version


def _event_to_schema_path(aggregate_cls, event_cls, avro_dir, version):
    aggregate_name = decode_cls_name(aggregate_cls)
    event_name = decode_cls_name(event_cls)

    filename = "v{version}_{aggregate_name}_{event_name}.json".format(
        aggregate_name=aggregate_name, event_name=event_name, version=version)

    return os.path.join(avro_dir, aggregate_name, filename)


def decode_cls_name(cls):
    """
    Convert camel case class name to snake case names used in event documentation.
    """
    return stringcase.snakecase(cls.__name__)


def parse_event_schema(spec_body):
    schema = avro.schema.Parse(spec_body)
    return schema


def get_schema_for_event(event_cls):
    if event_cls not in schemas:
        raise EventSchemaError("Cached Schema not found for: {}".format(event_cls))
    return schemas[event_cls]


def validate_event(event, schema=None):
    schema = schema or get_schema_for_event(event.__class__)
    return avro_validate(schema, event_to_json(event))



import warnings

from djangoevents.utils_abstract import abstract
from djangoevents.utils_abstract import is_abstract
from eventsourcing.domain.model.entity import DomainEvent
from eventsourcing.domain.model.entity import EventSourcedEntity


@abstract
class BaseAggregate(EventSourcedEntity):
    """
    `EventSourcedEntity` with saner mutator routing & naming:

    >>> class Asset(BaseAggregate):
    >>>    class Created(BaseAggregate.Created):
    >>>        def mutate(event, klass):
    >>>            return klass(...)
    >>>
    >>>    class Updated(DomainEvent):
    >>>        def mutate(event, instance):
    >>>            instance.connect = True
    >>>            return instance
    """

    @classmethod
    def is_abstract_class(cls):
        return is_abstract(cls)

    @classmethod
    def mutate(cls, aggregate=None, event=None):
        if aggregate:
            aggregate._validate_originator(event)

        if not hasattr(event, 'mutate_event'):
            msg = "{} does not provide a mutate_event() method.".format(event.__class__)
            raise NotImplementedError(msg)

        aggregate = event.mutate_event(event, aggregate or cls)
        aggregate._increment_version()
        return aggregate

    @classmethod
    def create_for_event(cls, event):
        aggregate = cls(
            entity_id=event.entity_id,
            domain_event_id=event.domain_event_id,
            entity_version=event.entity_version,
        )
        return aggregate


class BaseEntity(BaseAggregate):
    """
    `EventSourcedEntity` with saner mutator routing:

    OBSOLETE! Interface kept for backward compatibility.
    """
    @classmethod
    def is_abstract_class(cls):
        return super().is_abstract_class() or cls is BaseEntity

    @classmethod
    def mutate(cls, entity=None, event=None):
        warnings.warn("`BaseEntity` is depreciated. Please switch to: `BaseAggregate`", DeprecationWarning)
        return super().mutate(aggregate=entity, event=event)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-10-31 20:24
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('event_id', models.CharField(db_index=True, max_length=255)),
                ('event_type', models.CharField(max_length=255)),
                ('event_data', models.TextField()),
                ('aggregate_id', models.CharField(db_index=True, max_length=255)),
                ('aggregate_type', models.CharField(max_length=255)),
                ('aggregate_version', models.IntegerField()),
                ('create_date', models.DateTimeField()),
                ('metadata', models.TextField()),
                ('module_name', models.CharField(db_column='_module_name', max_length=255)),
                ('class_name', models.CharField(db_column='_class_name', max_length=255)),
                ('stored_entity_id', models.CharField(db_column='_stored_entity_id', max_length=255)),
            ],
            options={
                'db_table': 'event_journal',
            },
        ),
        migrations.AlterUniqueTogether(
            name='event',
            unique_together=set([('aggregate_id', 'aggregate_type', 'aggregate_version')]),
        ),
    ]



import avro.schema
from djangoevents.exceptions import EventSchemaError

import djangoevents.schema as schema
import json
import os
import pytest
import shutil
import tempfile

from djangoevents.domain import BaseAggregate
from djangoevents.domain import DomainEvent
from io import StringIO
from django.test import override_settings
from unittest import mock
from .test_domain import SampleEntity
from ..schema import set_event_version


class Project(BaseAggregate):
    class Created(BaseAggregate.Created):
        def mutate_event(self, event, klass):
            return klass(
                entity_id=event.entity_id,
                entity_version=event.entity_version,
                domain_event_id=event.domain_event_id,
                name=event.name,
            )

    class Closed(DomainEvent):
        # no mutate_event
        pass

    def __init__(self, project_id, name, **kwargs):
        super().__init__(**kwargs)
        self.project_id = project_id
        self.name = name


PROJECT_CREATED_SCHEMA = json.dumps({
    "name": "project_created",
    "type": "record",
    "doc": "Event posted when a project has been created",
    "namespace": "services.project_svc",
    "fields": [
        {
            "name": "entity_id",
            "type": "string",
            "doc": "ID of a the asset."
        },
        {
            "name": "entity_version",
            "type": "string",
            "doc": "Aggregate revision"
        },
        {
            "name": "domain_event_id",
            "type": "string",
            "doc": "ID of the last modifying event"
        },
        {
            "name": "name",
            "type": "string",
            "doc": "name of the project"
        }
    ]
})


@override_settings(BASE_DIR='/path/to/proj/src/')
@mock.patch.object(schema, 'list_concrete_aggregates', return_value=[Project])
@mock.patch.object(schema, 'load_event_schema')
def test_load_all_event_schemas(load_schema, load_aggs):
    schema.load_all_event_schemas()
    load_schema.assert_called_once_with(Project, Project.Created)


@override_settings(BASE_DIR='/path/to/proj/src/')
@mock.patch.object(schema, 'list_concrete_aggregates', return_value=[Project])
def test_load_all_event_schemas_missing_specs(list_aggs):
    with pytest.raises(EventSchemaError) as e:
        schema.load_all_event_schemas()

    path = "/path/to/proj/avro/project/v1_project_created.json"
    msg = "No event schema found for: {cls} (expecting file at:{path}).".format(cls=Project.Created, path=path)
    assert msg in str(e.value)


@override_settings(BASE_DIR='/path/to/proj/src/')
def test_valid_event_to_schema_path():
    SampleEntity.Created.version = None
    avro_path = schema.event_to_schema_path(aggregate_cls=SampleEntity, event_cls=SampleEntity.Created)
    assert avro_path == "/path/to/proj/avro/sample_entity/v1_sample_entity_created.json"


@override_settings(BASE_DIR='/path/to/proj/src/')
def test_event_to_schema_path_chooses_file_with_the_highest_version():
    try:
        # make temporary directory structure
        temp_dir = tempfile.mkdtemp()
        entity_dir = os.path.join(temp_dir, 'sample_entity')
        os.mkdir(entity_dir)

        for version in range(1, 4):
            # make empty schema file
            expected_schema_path = os.path.join(entity_dir, 'v{}_sample_entity_created.json'.format(version))
            with open(expected_schema_path, 'w'):
                pass

            # refresh version
            set_event_version(SampleEntity, SampleEntity.Created, avro_dir=temp_dir)

            # check path
            schema_path = schema.event_to_schema_path(
                aggregate_cls=SampleEntity,
                event_cls=SampleEntity.Created,
                avro_dir=temp_dir,
            )
            assert schema_path == expected_schema_path
    finally:
        # remove temporary directory
        shutil.rmtree(temp_dir)


def test_parse_invalid_event_schema():
    fp = StringIO("Invalid schema")

    with pytest.raises(avro.schema.SchemaParseException):
        schema.parse_event_schema(fp)


def test_parse_valid_event_schema():
    evt_schema = schema.parse_event_schema(PROJECT_CREATED_SCHEMA)
    assert evt_schema is not None


def test_validate_valid_event():
    class TestEvent:
        def __init__(self):
            self.entity_id = "1"
            self.entity_version = "XYZ"
            self.domain_event_id = "2"
            self.name = "Awesome Project"

    event = TestEvent()
    evt_schema = avro.schema.Parse(PROJECT_CREATED_SCHEMA)
    ret = schema.validate_event(event, evt_schema)
    assert ret is True


def test_validate_invalid_event():
    class TestEvent:
        def __init__(self):
            self.test = 'test'

    event = TestEvent()
    evt_schema = avro.schema.Parse(PROJECT_CREATED_SCHEMA)
    ret = schema.validate_event(event, evt_schema)
    assert ret is False


def test_decode_cls_name():
    class LongName:
        pass

    class Shortname:
        pass

    assert schema.decode_cls_name(LongName) == 'long_name'
    assert schema.decode_cls_name(Shortname) == 'shortname'



_abstract_classes = set()


def abstract(cls):
    """
    Decorator marking classes as abstract.

    The "abstract" mark is an internal tag.  Classes can be checked for
    being abstract with the `is_abstract` function.  The tag is non
    inheritable: every class or subclass has to be explicitly marked
    with the decorator to be considered abstract.
    """

    _abstract_classes.add(cls)
    return cls


def is_abstract(cls):
    return cls in _abstract_classes



from django.http import Http404
import warnings


def get_aggregate_or_404(repo, aggregate_id):
    """
    Uses `repo` to return most recent state of the aggregate. Raises a Http404 exception if the
    `aggregate_id` does not exist.

    `repo` has to be a EventSourcedRepository object.
    """
    if aggregate_id not in repo:
        raise Http404
    return repo[aggregate_id]


def get_entity_or_404(repo, entity_id):
    warnings.warn("`get_entity_or_404` is depreciated. Please switch to: `get_aggregate_or_404`", DeprecationWarning)
    return get_aggregate_or_404(repo, entity_id)



from setuptools import setup
from setuptools.command.upload import upload
import os


class ReleaseToPyPICommand(upload):

    def finalize_options(self):
        self.repository = 'https://upload.pypi.org/legacy/'
        self.username = os.environ['PYPI_USERNAME']
        self.password = os.environ['PYPI_PASSWORD']


setup(
    name='djangoevents',
    version='0.13.3',
    url='https://github.com/ApplauseOSS/djangoevents',
    license='MIT',
    description='Building blocks for building Event Sourcing Django applications.',
    author='Applause',
    author_email='eng@applause.com',
    zip_safe=False,
    packages=[
        'djangoevents',
        'djangoevents.migrations',
        'djangoevents.tests.settings',
    ],
    include_package_data=True,
    install_requires=[
        'eventsourcing>=1.2,<1.3',
        'django',
        'avro-python3==1.7.7',
        'stringcase==1.0.6',
    ],
    cmdclass={
        'release_to_pypi': ReleaseToPyPICommand
    }
)



"""
Parses event schemas and reports on their validity.
"""
from django.core.management.base import BaseCommand
from djangoevents.exceptions import EventSchemaError
from djangoevents.schema import load_all_event_schemas, schemas
from djangoevents.settings import is_validation_enabled


class Command(BaseCommand):

    def handle(self, *args, **options):
        print("=> Searching for schema files...")
        try:
            if is_validation_enabled():
                print("--> Schema validation enabled. Checking state..")
            load_all_event_schemas()
        except EventSchemaError as e:
            print("Missing or invalid event schemas:")
            print(e)
        else:
            print("--> Detected events:")
            for item in schemas.keys():
                print("--> {}".format(item))
            print("--> All schemas valid!")
        print("=> Done.")



import warnings

from .repository import DjangoStoredEventRepository
from .unifiedtranscoder import UnifiedTranscoder
from django.core.serializers.json import DjangoJSONEncoder
from eventsourcing.application.base import EventSourcingApplication
from eventsourcing.infrastructure.event_sourced_repo import EventSourcedRepository


class EventSourcingWithDjango(EventSourcingApplication):
    def __init__(self, **kwargs):
        kwargs.setdefault('json_encoder_cls', DjangoJSONEncoder)
        super().__init__(**kwargs)
        self.on_init()

    def on_init(self):
        """
        Override at subclass. For example, initialize repositories that this
        app would use.
        """

    def create_stored_event_repo(self, **kwargs):
        return DjangoStoredEventRepository(**kwargs)

    def create_transcoder(self, always_encrypt, cipher, json_decoder_cls, json_encoder_cls):
        return UnifiedTranscoder(json_encoder_cls=json_encoder_cls)

    def get_repo_for_aggregate(self, aggregate_cls):
        """
        Returns EventSourcedRepository class for a given aggregate.
        """
        clsname = '%sRepository' % aggregate_cls.__name__
        repo_cls = type(clsname, (EventSourcedRepository,), {'domain_class': aggregate_cls})
        return repo_cls(event_store=self.event_store)

    def get_repo_for_entity(self, entity_cls):
        """
        Returns EventSourcedRepository class for given entity.
        """
        msg = "`get_repo_for_entity` is depreciated. Please switch to: `get_repo_for_aggregate`"
        warnings.warn(msg, DeprecationWarning)
        return self.get_repo_for_aggregate(aggregate_cls=entity_cls)



from .domain import DomainEvent
from .schema import get_event_version
from .utils import camel_case_to_snake_case
from collections import namedtuple
from datetime import datetime
from eventsourcing.domain.model.events import resolve_attr
from eventsourcing.domain.services.transcoding import AbstractTranscoder
from eventsourcing.domain.services.transcoding import ObjectJSONDecoder
from eventsourcing.domain.services.transcoding import id_prefix_from_event
from eventsourcing.domain.services.transcoding import make_stored_entity_id
from eventsourcing.utils.time import timestamp_from_uuid
from inspect import isclass

import importlib
import json


UnifiedStoredEvent = namedtuple('UnifiedStoredEvent', [
    'event_id',
    'event_type',
    'event_version',
    'event_data',
    'aggregate_id',
    'aggregate_type',
    'aggregate_version',
    'create_date',
    'metadata',
    'module_name',
    'class_name',
    'stored_entity_id',
])


class UnifiedTranscoder(AbstractTranscoder):
    def __init__(self, json_encoder_cls=None):
        self.json_encoder_cls = json_encoder_cls
        # encrypt not implemented

    def serialize(self, domain_event):
        """
           Serializes a domain event into a stored event.
        """
        assert isinstance(domain_event, DomainEvent)

        event_data = {key: value for key, value in domain_event.__dict__.items() if key not in {
            'domain_event_id',
            'entity_id',
            'entity_version',
            'metadata',
        }}

        domain_event_class = type(domain_event)
        event_version = get_event_version(domain_event_class)

        return UnifiedStoredEvent(
            event_id=domain_event.domain_event_id,
            event_type=get_event_type(domain_event),
            event_version=event_version,
            event_data=self._json_encode(event_data),
            aggregate_id=domain_event.entity_id,
            aggregate_type=get_aggregate_type(domain_event),
            aggregate_version=domain_event.entity_version,
            create_date=datetime.fromtimestamp(timestamp_from_uuid(domain_event.domain_event_id)),
            metadata=self._json_encode(getattr(domain_event, 'metadata', None)),
            module_name=domain_event_class.__module__,
            class_name=domain_event_class.__qualname__,
            # have to have stored_entity_id because of the lib
            stored_entity_id=make_stored_entity_id(id_prefix_from_event(domain_event), domain_event.entity_id),
        )

    def deserialize(self, stored_event):
        """
        Recreates original domain event from stored event topic and event attrs.
        """
        assert isinstance(stored_event, UnifiedStoredEvent)
        # Get the domain event class from the topic.
        domain_event_class = self._get_domain_event_class(stored_event.module_name, stored_event.class_name)

        # Deserialize event attributes from JSON
        event_attrs = self._json_decode(stored_event.event_data)

        # Reinstantiate and return the domain event object.
        try:
            domain_event = domain_event_class(entity_id=stored_event.aggregate_id,
                                              entity_version=stored_event.aggregate_version,
                                              domain_event_id=stored_event.event_id,
                                              **event_attrs)
        except TypeError:
            raise ValueError("Unable to instantiate class '{}' with data '{}'"
                             "".format(stored_event.class_name, event_attrs))

        return domain_event

    @staticmethod
    def _get_domain_event_class(module_name, class_name):
        """Return domain class described by given topic.

        Args:
            module_name: string of module_name
            class_name: string of class_name
        Returns:
            A domain class.

        Raises:
            ResolveDomainFailed: If there is no such domain class.
        """
        try:
            module = importlib.import_module(module_name)
        except ImportError as e:
            raise ResolveDomainFailed("{}#{}: {}".format(module_name, class_name, e))
        try:
            domain_event_class = resolve_attr(module, class_name)
        except AttributeError as e:
            raise ResolveDomainFailed("{}#{}: {}".format(module_name, class_name, e))

        if not isclass(domain_event_class):
            raise ValueError("Event class is not a type: {}".format(domain_event_class))

        if not issubclass(domain_event_class, DomainEvent):
            raise ValueError("Event class is not a DomainEvent: {}".format(domain_event_class))

        return domain_event_class

    def _json_encode(self, data):
        return json.dumps(data, separators=(',', ':'), sort_keys=True, cls=self.json_encoder_cls)

    def _json_decode(self, json_str):
        return json.loads(json_str, cls=ObjectJSONDecoder)


class ResolveDomainFailed(Exception):
    pass


def get_aggregate_type(domain_event):
    assert isinstance(domain_event, DomainEvent)
    domain_event_class = type(domain_event)
    return domain_event_class.__qualname__.split('.')[0]


def get_event_type(domain_event):
    if hasattr(domain_event, 'event_type'):
        return getattr(domain_event, 'event_type')
    else:
        aggregate_type = get_aggregate_type(domain_event)
        event_name = domain_event.__class__.__name__
        event_type_name = '%s%s' % (aggregate_type, event_name)
        return camel_case_to_snake_case(event_type_name)
